/*******************************************************************************
 * Copyright (c) 2000, 2004 IBM Corporation and others.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * Contributors:
 *     IBM Corporation - initial API and implementation
 *******************************************************************************/
package org.eclipse.target.internal.iofile;

import java.io.*;
import java.util.*;

import org.eclipse.core.runtime.*;
import org.eclipse.target.*;
import org.eclipse.target.internal.Utils;

public class IoFileTargetResource extends TargetResource {

	private final String siteRelativePath;
	private final File file;

	public IoFileTargetResource(Site site, File file, String siteRelativePath) {
		super(site);
		this.file = file;
		this.siteRelativePath = siteRelativePath;
	}

	/* (non-Javadoc)
	 * @see org.eclipse.target.ITargetResource#getRelativePath()
	 */
	public String getRelativePath() {
		return siteRelativePath;
	}

	/* (non-Javadoc)
	 * @see org.eclipse.target.ITargetResource#isDirectory()
	 */
	public boolean isDirectory() {
		return asFile().isDirectory();
	}

	private File asFile() {
		return file;
	}

	/* (non-Javadoc)
	 * @see org.eclipse.target.ITargetResource#exists(org.eclipse.core.runtime.IProgressMonitor)
	 */
	public boolean exists(IProgressMonitor monitor) throws CoreException {
		return asFile().exists();
	}

	/* (non-Javadoc)
	 * @see org.eclipse.target.ITargetResource#members(org.eclipse.core.runtime.IProgressMonitor)
	 */
	public ITargetResource[] members(IProgressMonitor monitor) throws CoreException {
		File[] files = asFile().listFiles();
		List result = new ArrayList(files.length);
		for (int i = 0; i < files.length; i++) {
			File file = files[i];
			result.add(new IoFileTargetResource(getSite(), file, Utils.appendPath(getRelativePath(), file.getName())));
		}
		return (ITargetResource[]) result.toArray(new ITargetResource[result.size()]);
	}

	/* (non-Javadoc)
	 * @see org.eclipse.target.ITargetResource#getContents(org.eclipse.core.runtime.IProgressMonitor)
	 */
	public InputStream getContents(IProgressMonitor progress) throws CoreException {
		try {
			return new FileInputStream(asFile());
		} catch (FileNotFoundException e) {
			throw IoFilePlugin.wrapException(e);
		}
	}

	/* (non-Javadoc)
	 * @see org.eclipse.target.ITargetResource#setContents(java.io.InputStream, long, org.eclipse.core.runtime.IProgressMonitor)
	 */
	public void setContents(InputStream stream, long size, IProgressMonitor progress) throws CoreException {
		OutputStream os = null;
		try {
			os = new BufferedOutputStream(new FileOutputStream(asFile()));
			byte[] buffer = new byte[1024];
			int count;
			while ((count = stream.read(buffer)) != -1) {
				os.write(buffer, 0, count);
			}
		} catch (IOException e) {
			throw IoFilePlugin.wrapException(e);
		} finally {
			try {
				stream.close();
			} catch (IOException e1) {
				// Ignore
			} finally {
				try {
					if (os != null) os.close();
				} catch (IOException e2) {
					// Ignore
				}
			}
		}

	}

	/* (non-Javadoc)
	 * @see org.eclipse.target.ITargetResource#delete(org.eclipse.core.runtime.IProgressMonitor)
	 */
	public void delete(IProgressMonitor monitor) throws CoreException {
		if (!asFile().delete()) {
			throw new CoreException(new Status(IStatus.ERROR, IoFilePlugin.ID, ITargetStatus.DELETION_FAILED, "Failed to delete file " + asFile().getAbsolutePath(), null));
		}
	}

	/* (non-Javadoc)
	 * @see org.eclipse.target.ITargetResource#fetchProperties(org.eclipse.core.runtime.QualifiedName[], boolean, org.eclipse.core.runtime.IProgressMonitor)
	 */
	public QualifiedName[] fetchProperties(QualifiedName[] properties, boolean refresh, IProgressMonitor monitor) throws CoreException {
		return properties;
	}

	/* (non-Javadoc)
	 * @see org.eclipse.target.ITargetResource#isAvailable(org.eclipse.core.runtime.QualifiedName)
	 */
	public boolean isAvailable(QualifiedName property) {
		return isSupported(property);
	}

	/* (non-Javadoc)
	 * @see org.eclipse.target.ITargetResource#getProperty(org.eclipse.core.runtime.QualifiedName)
	 */
	public String getProperty(QualifiedName property) throws CoreException {
		if (property.equals(SIZE_PROPERTY)) {
			return Long.toString(getSize());
		}
		if (property.equals(LAST_MODIFIED_TIMESTAMP_PROPERTY)) {
			return new Date(asFile().lastModified()).toString();
		}
		if (property.equals(CONTENT_IDENTIFIER_PROPERTY)) {
			return new Date(asFile().lastModified()).toString() + getSize();
		}
		return null;
	}

	/* (non-Javadoc)
	 * @see org.eclipse.target.ITargetResource#getSize()
	 */
	public long getSize() {
		return asFile().length();
	}

	/* (non-Javadoc)
	 * @see org.eclipse.target.ITargetResource#mkdir(org.eclipse.core.runtime.IProgressMonitor)
	 */
	public void mkdir(IProgressMonitor monitor) throws CoreException {
		if (!asFile().mkdir()) {
			throw new CoreException(new Status(IStatus.ERROR, IoFilePlugin.ID, ITargetStatus.CREATION_FAILED, "Failed to create folder " + asFile().getAbsolutePath(), null));
		}
	}

	/* (non-Javadoc)
	 * @see org.eclipse.target.ITargetResource#mkdirs(org.eclipse.core.runtime.IProgressMonitor)
	 */
	public void mkdirs(IProgressMonitor monitor) throws CoreException {
		if (!asFile().mkdirs()) {
			throw new CoreException(new Status(IStatus.ERROR, IoFilePlugin.ID, ITargetStatus.CREATION_FAILED, "Failed to create folder " + asFile().getAbsolutePath(), null));
		}
	}

	/* (non-Javadoc)
	 * @see org.eclipse.target.TargetResource#createResource(java.lang.String, boolean)
	 */
	protected ITargetResource createResource(String relativePath, boolean isDirectory) {
		File resourceFile;
		String resourcePath;
		if (relativePath.equals(PARENT_RELATIVE_PATH)) {
			resourcePath = Utils.removeLastSegment(siteRelativePath);
			resourceFile = asFile().getParentFile();
		} else {
			resourcePath = Utils.appendPath(getRelativePath(), relativePath);
			resourceFile = new File(asFile(), convertToFileSystemPath(relativePath));
		}
		return new IoFileTargetResource(getSite(), resourceFile, resourcePath);
	}

	private String convertToFileSystemPath(String relativePath) {
		if (File.separatorChar != '/') {
			return relativePath.replace('/', File.separatorChar);
		} else {
			return relativePath;
		}
	}
	
	/* (non-Javadoc)
	 * @see org.eclipse.target.TargetResource#getFullPath()
	 */
	public String getFullPath() {
		return convertToFileSystemPath(super.getFullPath());
	}

}
