/*******************************************************************************
 * Copyright (c) 2000, 2004 IBM Corporation and others.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * Contributors:
 *     IBM Corporation - initial API and implementation
 *******************************************************************************/
package org.eclipse.target.internal.ui;

import java.util.ArrayList;
import java.util.List;

import org.eclipse.core.runtime.*;
import org.eclipse.core.runtime.jobs.ISchedulingRule;
import org.eclipse.jface.resource.ImageDescriptor;
import org.eclipse.target.ITargetResource;
import org.eclipse.target.internal.TargetPlugin;
import org.eclipse.target.internal.Utils;
import org.eclipse.ui.ISharedImages;
import org.eclipse.ui.PlatformUI;
import org.eclipse.ui.model.IWorkbenchAdapter;
import org.eclipse.ui.progress.IDeferredWorkbenchAdapter;
import org.eclipse.ui.progress.IElementCollector;

/**
 * Used to show IRemoteTargetResource instances in the UI. In addition these elements
 * support caching of values returned from IRemoteTargetResource methods, as such these
 * instances aren't as much proxies as the underlying remote handles.
 * <p>
 * Implementation in progress: support for configuring these elements with a custom
 * progress monitor that can be the target of long running operations. This will
 * allow showing progress in wizards and in the SiteExplorerView when navigating
 * remote servers.</p>
 */
public class TargetResourceElement implements IDeferredWorkbenchAdapter, IWorkbenchAdapter, IAdaptable {
	
	public TargetResourceElement() {
	}
	
	public ITargetResource getRemoteResource(Object object) {
		if (object instanceof ITargetResource) {
			return (ITargetResource)object;
		}
		return null;
	}
	
	public Object getAdapter(Class adapter) {
		if (adapter == IWorkbenchAdapter.class) return this;
		return null;
	}

	public Object[] getChildren(Object o) {
		ITargetResource remote = getRemoteResource(o);
		if (remote == null) {
			return new Object[0];
		}
		try {
			return internalFetchChildren(remote, new NullProgressMonitor());
		} catch (CoreException e) {
			Utils.handle(null, e);
			return new Object[0];
		}
	}
	
	public ImageDescriptor getImageDescriptor(Object object) {
		ITargetResource remote = getRemoteResource(object);
		if (remote == null) {
			return null;
		}
		if(remote.isDirectory()) {
			return PlatformUI.getWorkbench().getSharedImages().getImageDescriptor(ISharedImages.IMG_OBJ_FOLDER);
		} else {
			return PlatformUI.getWorkbench().getEditorRegistry().getImageDescriptor(remote.getName());
		}
	}
	
	public String getLabel(Object o) {
		ITargetResource remote = getRemoteResource(o);
		if (remote == null) {
			return null;
		}
		return remote.getName();
	}
	
	public Object getParent(Object o) {
		ITargetResource remote = getRemoteResource(o);
		if (remote == null) {
			return null;
		}
		return remote.getParent();
	}

	/* (non-Javadoc)
	 * @see org.eclipse.ui.progress.IDeferredWorkbenchAdapter#fetchDeferredChildren(java.lang.Object, org.eclipse.ui.progress.IElementCollector, org.eclipse.core.runtime.IProgressMonitor)
	 */
	public void fetchDeferredChildren(Object object, IElementCollector collector, IProgressMonitor monitor) {
		try {
			monitor.beginTask(null, IProgressMonitor.UNKNOWN);
			ITargetResource remote = getRemoteResource(object);
			if (remote != null) {
				collector.add(
						internalFetchChildren(remote, TargetPlugin.subMonitorFor(monitor, 95)), 
						TargetPlugin.subMonitorFor(monitor, 5));
			}
			collector.done();
		} catch (CoreException e) {
			Utils.handle(null, e);
		} finally {
			monitor.done();
		}
	}

	private ITargetResource[] internalFetchChildren(ITargetResource remote, IProgressMonitor monitor) throws CoreException {
		try {
			monitor.beginTask(null, IProgressMonitor.UNKNOWN);
			ITargetResource[] members = remote.members(TargetPlugin.subMonitorFor(monitor, 50));
			List remoteElements = new ArrayList();
			for (int i = 0; i < members.length; i++) {
				ITargetResource child = members[i];
						
				// cache size and last modified
				child.fetchProperties(new QualifiedName[] {
						ITargetResource.SIZE_PROPERTY,
						ITargetResource.LAST_MODIFIED_TIMESTAMP_PROPERTY,
				}, false, TargetPlugin.subMonitorFor(monitor, 25));						
				remoteElements.add(child);
			}
			return (ITargetResource[]) remoteElements.toArray(new ITargetResource[remoteElements.size()]);
		} finally {
			monitor.done();
		}													
	}

	/* (non-Javadoc)
	 * @see org.eclipse.ui.progress.IDeferredWorkbenchAdapter#isContainer()
	 */
	public boolean isContainer() {
		return true;
	}

	/* (non-Javadoc)
	 * @see org.eclipse.ui.progress.IDeferredWorkbenchAdapter#getRule(java.lang.Object)
	 */
	public ISchedulingRule getRule(Object object) {
		return null;
	}
}
