/*******************************************************************************
 * Copyright (c) 2000, 2004 IBM Corporation and others.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * Contributors:
 *     IBM Corporation - initial API and implementation
 *******************************************************************************/
package org.eclipse.target.internal.webdav;

import java.io.IOException;
import java.io.InputStream;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.*;

import org.eclipse.core.runtime.*;
import org.eclipse.target.*;
import org.eclipse.target.internal.Utils;
import org.eclipse.target.streams.PollingInputStream;
import org.eclipse.target.streams.ProgressMonitorInputStream;
import org.eclipse.webdav.*;
import org.eclipse.webdav.client.*;
import org.eclipse.webdav.dom.ElementEditor;
import org.eclipse.webdav.dom.MalformedElementException;
import org.eclipse.webdav.internal.kernel.*;

public class WebDavTargetResource extends TargetResource {

	public static final int TRANSFER_PROGRESS_INCREMENT = 32768;
	
	/*
	 * The WebDAVFactory is used to create new resource handles, locators, and
	 * contexts, etc.
	 */
	protected static WebDAVFactory DAV_FACTORY = new WebDAVFactory();
	
	private final String siteRelativePath;
	private final boolean isDirectory;
	private AbstractResourceHandle remoteResource;
	
	private Map propertyCache = new HashMap();

	public static ITargetResource forSite(WebDavSite site, DAVClient davClient) {
		return createChild(site, davClient, "", true); //$NON-NLS-1$
	}
	public static ResourceHandle createResource(WebDavSite site, DAVClient davClient, URL url) {
		ILocator locator = DAV_FACTORY.newLocator(url.toExternalForm());
		ResourceHandle davHandle = new ResourceHandle(davClient, locator);
		return davHandle;
	}
	
	private static ITargetResource createChild(WebDavSite site, DAVClient client, String path, boolean isContainer) {
		String fullPath = Utils.appendPath(site.getLocation(), path);
		if(isContainer)
			fullPath = Utils.makeAbsolute(fullPath);
			
		ILocator locator = DAV_FACTORY.newLocator(fullPath);
		AbstractResourceHandle handle = new ResourceHandle(client, locator);
		return new WebDavTargetResource(site, handle, path, isContainer);
	}
	
	public WebDavTargetResource(Site site, AbstractResourceHandle remoteResource, String siteRelativePath, boolean isDirectory) {
		super(site);
		this.siteRelativePath = siteRelativePath;
		this.isDirectory = isDirectory;
		this.remoteResource = remoteResource;
	}

	/* (non-Javadoc)
	 * @see org.eclipse.target.ITargetResource#getRelativePath()
	 */
	public String getRelativePath() {
		return siteRelativePath;
	}

	/* (non-Javadoc)
	 * @see org.eclipse.target.ITargetResource#isDirectory()
	 */
	public boolean isDirectory() {
		return isDirectory;
	}

	/* (non-Javadoc)
	 * @see org.eclipse.target.ITargetResource#exists(org.eclipse.core.runtime.IProgressMonitor)
	 */
	public boolean exists(IProgressMonitor monitor) throws CoreException {
		try {
			return remoteResource.exists();
		} catch (DAVException e) {
			throw WebdavPlugin.wrapException(e);
		}
	}

	/* (non-Javadoc)
	 * @see org.eclipse.target.ITargetResource#members(org.eclipse.core.runtime.IProgressMonitor)
	 */
	public ITargetResource[] members(IProgressMonitor monitor) throws CoreException {
		try {		
			monitor.beginTask(Policy.bind("DavRemoteTargetResource.membersProgress", getName()), 100); //$NON-NLS-1$
			monitor.subTask(Policy.bind("DavRemoteTargetResource.membersProgress", getName())); //$NON-NLS-1$
			// Only containers have children.
			if (!isDirectory())
				return new ITargetResource[0];
	
			// Query the DAV server to find the children (i.e., members) of the receiver.
			Set remoteChildren;
			try {
				remoteChildren = remoteResource.asCollectionHandle().getMembers();
				monitor.worked(80);
			} catch (DAVException exception) {
				// Problem getting the children.
				throw WebdavPlugin.wrapException(exception);
			}
	
			// Have to contruct resource states from DAV resource handles.
			WebDavTargetResource[] result = new WebDavTargetResource[remoteChildren.size()];
			int index = 0;
			Iterator remoteChildrenItr = remoteChildren.iterator();
			while (remoteChildrenItr.hasNext()) {
				AbstractResourceHandle remoteChild = (AbstractResourceHandle)remoteChildrenItr.next();
				result[index++] = new WebDavTargetResource(getSite(), remoteChild, Utils.appendPath(siteRelativePath, Utils.getLastSegment(remoteChild.getLocator().getResourceURL())), (remoteChild instanceof CollectionHandle));
			}
			monitor.worked(20);
	
			// All children processed.
			return result;
		} finally {
			monitor.done();
		}
	}

	/* (non-Javadoc)
	 * @see org.eclipse.target.ITargetResource#getContents(org.eclipse.core.runtime.IProgressMonitor)
	 */
	public InputStream getContents(IProgressMonitor progress) throws CoreException {
		try {
			// number of bytes expected
			final long size = getSize() >> 10;
			final String name = Utils.toTruncatedPath(getFullPath(), 3);
			
			// timeout value for this site
			int timeout = ((WebDavSite)getSite()).getTimeout();
			
			final String prefix = Policy.bind("DavRemoteTargetResource.download", name); //$NON-NLS-1$
			progress.subTask(Policy.bind("DavRemoteTargetResource.transferNoSize", prefix)); //$NON-NLS-1$
			
			// wrap the dav input stream with an input stream that allows progress
			// and cancellation.
			InputStream in = new ProgressMonitorInputStream(
									new PollingInputStream(remoteResource.getContent(), 
									timeout,
									progress), 
								0, 
								TRANSFER_PROGRESS_INCREMENT, 
								progress) {
				// perform updates to the progress message, showing the number of bytes transfered
				protected void updateMonitor(long bytesRead, long bytesTotal, IProgressMonitor monitor) {
					if (bytesRead == 0) return;
					monitor.subTask(Policy.bind("DavRemoteTargetResource.transfer", //$NON-NLS-1$
						new Object[] { prefix,
										Long.toString(bytesRead >> 10),
										Long.toString(size) }));
				}
			};																			
											
			return in;
		} catch (ClientException exception) {
			if (exception.getStatusCode() == IResponse.SC_NOT_FOUND)
				throw new CoreException(new org.eclipse.core.runtime.Status(IStatus.ERROR, WebdavPlugin.ID, ITargetStatus.RESOURCE_NOT_AVAILABLE, exception.getMessage(), exception));
			throw WebdavPlugin.wrapException(exception);
		} catch (DAVException exception) {
			throw WebdavPlugin.wrapException(exception);
		}
	}

	/* (non-Javadoc)
	 * @see org.eclipse.target.ITargetResource#setContents(java.io.InputStream, long, org.eclipse.core.runtime.IProgressMonitor)
	 */
	public void setContents(InputStream is, final long size, IProgressMonitor progress) throws CoreException {
		// number of bytes expected
		final String name = Utils.toTruncatedPath(getFullPath(), 3);
		
		// timeout value for this site
		int timeout = ((WebDavSite)getSite()).getTimeout();

		final String prefix = Policy.bind("DavRemoteTargetResource.upload", name); //$NON-NLS-1$
		progress.subTask(Policy.bind("DavRemoteTargetResource.transferNoSize", prefix)); //$NON-NLS-1$
				
		// wrap the dav input stream with an input stream that allows progress
		// and cancellation.
		InputStream in = new ProgressMonitorInputStream(
								new PollingInputStream(is, 
								timeout,
								progress), 
							0, 
							TRANSFER_PROGRESS_INCREMENT, 
							progress) {
			// perform updates to the progress message, showing the number of bytes transfered
			protected void updateMonitor(long bytesRead, long bytesTotal, IProgressMonitor monitor) {
				if (bytesRead == 0) return;
				monitor.subTask(Policy.bind("DavRemoteTargetResource.transfer", //$NON-NLS-1$
					new Object[] { prefix, Long.toString(bytesRead >> 10), Long.toString(size >> 10) }));
			}
		};
		boolean success = false;
		try {
			remoteResource.setContent(in);
			success = true;
		} catch (DAVException e) {
			throw WebdavPlugin.wrapException(e);
		} finally {
			if (!success) {
				try {
					in.close();
				} catch (IOException e1) {
					// Ignore
				}
			}
		}
	}

	/* (non-Javadoc)
	 * @see org.eclipse.target.ITargetResource#delete(org.eclipse.core.runtime.IProgressMonitor)
	 */
	public void delete(IProgressMonitor monitor) throws CoreException {
		try {
			remoteResource.delete();
		} catch (WebDAVException exception) {
			throw WebdavPlugin.wrapException(exception);
		} catch (DAVException exception) {
			throw WebdavPlugin.wrapException(exception);
		}
	}

	/* (non-Javadoc)
	 * @see org.eclipse.target.ITargetResource#fetchProperties(org.eclipse.core.runtime.QualifiedName[], boolean, org.eclipse.core.runtime.IProgressMonitor)
	 */
	public org.eclipse.core.runtime.QualifiedName[] fetchProperties(QualifiedName[] properties,
			boolean refresh, IProgressMonitor monitor) throws CoreException {
		propertyCache.clear();
		Collection davProperties = getDavProperties(properties);
		List fetchedProperties = new ArrayList();
		try {
			URLTable table = remoteResource.getProperties(davProperties, IContext.DEPTH_ZERO);
			Map result = (Map)table.get(new URL(getFullPath()));
            if (result != null) {
    			for (Iterator iter = result.values().iterator(); iter.hasNext();) {
    				PropertyStatus status = (PropertyStatus) iter.next();
    				if (status.getStatusCode() == IResponse.SC_OK) {
    					org.eclipse.core.runtime.QualifiedName targetProperty = getTargetProperty(status.getPropertyName());
    					if (targetProperty != null) {
    						propertyCache.put(targetProperty, status);
    						fetchedProperties.add(targetProperty);
    					}
    				}
    			}
            }
		} catch (DAVException e) {
			throw WebdavPlugin.wrapException(e);
		} catch (MalformedURLException e) {
			throw WebdavPlugin.wrapException(e);
		} catch (MalformedElementException e) {
			throw WebdavPlugin.wrapException(e);
		}
		return (org.eclipse.core.runtime.QualifiedName[]) fetchedProperties.toArray(new org.eclipse.core.runtime.QualifiedName[fetchedProperties.size()]);
	}

	private org.eclipse.core.runtime.QualifiedName getTargetProperty(org.eclipse.webdav.dom.QualifiedName name) {
		if (name.equals(WebDAVPropertyNames.DAV_GET_LAST_MODIFIED)) {
			return LAST_MODIFIED_TIMESTAMP_PROPERTY;
		}
		if (name.equals(WebDAVPropertyNames.DAV_GET_E_TAG)) {
			return CONTENT_IDENTIFIER_PROPERTY;
		}
		if (name.equals(WebDAVPropertyNames.DAV_GET_CONTENT_LENGTH)) {
			return SIZE_PROPERTY;
		}
		if (name.equals(WebDAVPropertyNames.DAV_CREATOR_DISPLAYNAME)) {
			return CREATOR_DISPLAY_NAME_PROPERTY;
		}
		return null;
	}

	private org.eclipse.webdav.dom.QualifiedName getDavProperty(org.eclipse.core.runtime.QualifiedName name) {
		if (name.equals(LAST_MODIFIED_TIMESTAMP_PROPERTY)) {
			return (WebDAVPropertyNames.DAV_GET_LAST_MODIFIED);
		} else if (name.equals(CONTENT_IDENTIFIER_PROPERTY)) {
			return (WebDAVPropertyNames.DAV_GET_E_TAG);
		} else if (name.equals(SIZE_PROPERTY)) {
			return (WebDAVPropertyNames.DAV_GET_CONTENT_LENGTH);
		} else if (name.equals(CREATOR_DISPLAY_NAME_PROPERTY)) {
			return (WebDAVPropertyNames.DAV_CREATOR_DISPLAYNAME);
		}
		return null;
	}
	
	private Collection getDavProperties(org.eclipse.core.runtime.QualifiedName[] properties) {
		ArrayList result = new ArrayList();
		for (int i = 0; i < properties.length; i++) {
			org.eclipse.webdav.dom.QualifiedName davName = getDavProperty(properties[i]);
			if (davName != null) {
				result.add(davName);
			}
		}
		return result;
	}

	/* (non-Javadoc)
	 * @see org.eclipse.target.ITargetResource#isAvailable(org.eclipse.core.runtime.QualifiedName)
	 */
	public boolean isAvailable(org.eclipse.core.runtime.QualifiedName property) {
		return propertyCache.keySet().contains(property);
	}

	/* (non-Javadoc)
	 * @see org.eclipse.target.ITargetResource#getSupportedProperties()
	 */
	public org.eclipse.core.runtime.QualifiedName[] getSupportedProperties() {
		return new QualifiedName[] { LAST_MODIFIED_TIMESTAMP_PROPERTY, SIZE_PROPERTY, CONTENT_IDENTIFIER_PROPERTY };
	}

	/* (non-Javadoc)
	 * @see org.eclipse.target.ITargetResource#getProperty(org.eclipse.core.runtime.QualifiedName)
	 */
	public String getProperty(QualifiedName property) throws CoreException {
		PropertyStatus status = (PropertyStatus)propertyCache.get(property);
		if (status == null) return null;
		return ElementEditor.getFirstText(status.getProperty());
	}

	/* (non-Javadoc)
	 * @see org.eclipse.target.ITargetResource#getSize()
	 */
	public long getSize() {
		try {
			String sizeString = getProperty(SIZE_PROPERTY);
			if (sizeString != null) {
				return Long.parseLong(sizeString);
			}
		} catch (CoreException e) {
			WebdavPlugin.log(e);
		}
		return 0;
	}

	/* (non-Javadoc)
	 * @see org.eclipse.target.TargetResource#createResource(java.lang.String, boolean)
	 */
	protected ITargetResource createResource(String relativePath, boolean isDirectory) {
		String resourcePath;
		if (relativePath.equals(PARENT_RELATIVE_PATH)) {
			resourcePath = Utils.removeLastSegment(siteRelativePath);
		} else {
			resourcePath = Utils.appendPath(getRelativePath(), relativePath);
		}
		return createChild(((WebDavSite)getSite()), remoteResource.getDAVClient(), resourcePath, isDirectory);
	}

	/* (non-Javadoc)
	 * @see org.eclipse.target.ITargetResource#mkdir(org.eclipse.core.runtime.IProgressMonitor)
	 */
	public void mkdir(IProgressMonitor monitor) throws CoreException {
		try {
			remoteResource.asCollectionHandle().create();
		} catch (DAVException e) {
			throw WebdavPlugin.wrapException(e);
		}
	}

	/* (non-Javadoc)
	 * @see org.eclipse.target.ITargetResource#mkdirs(org.eclipse.core.runtime.IProgressMonitor)
	 */
	public void mkdirs(IProgressMonitor monitor) throws CoreException {
		try {
			remoteResource.asCollectionHandle().mkdirs();
		} catch (DAVException e) {
			throw WebdavPlugin.wrapException(e);
		}
	}
	
	public boolean canBeReached() throws CoreException {
		try {
			return this.remoteResource.canTalkDAV();
		} catch (DAVException e) {
			throw WebdavPlugin.wrapException(e);
		}
	}

}
