/*******************************************************************************
 * Copyright (c) 2000, 2004 IBM Corporation and others.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * Contributors:
 *     IBM Corporation - initial API and implementation
 *******************************************************************************/
package org.eclipse.target.ui;

import java.lang.reflect.InvocationTargetException;

import org.eclipse.core.runtime.*;
import org.eclipse.jface.dialogs.MessageDialog;
import org.eclipse.jface.operation.IRunnableWithProgress;
import org.eclipse.jface.util.Assert;
import org.eclipse.jface.wizard.IWizardContainer;
import org.eclipse.jface.wizard.Wizard;
import org.eclipse.osgi.util.NLS;
import org.eclipse.target.*;
import org.eclipse.target.internal.TargetPlugin;

/**
 * Handle common behavior for a simple one-page site creaton wizard.
 * <p>
 * <b>Note:</b> This class/interface is part of an interim API that is still under 
 * development and expected to change significantly before reaching stability. 
 * It is being made available at this early stage to solicit feedback from pioneering 
 * adopters on the understanding that any code that uses this API will almost 
 * certainly be broken (repeatedly) as the API evolves.
 * </p> 
 */
public abstract class SiteCreationWizard extends Wizard {

	private SiteCreationWizardPage mainPage;

	/* (non-Javadoc)
	 * @see org.eclipse.jface.wizard.IWizard#addPages()
	 */
	public void addPages() {
		mainPage = createSitePage();
		if (mainPage != null)
			addPage(mainPage);
	}
	
	/**
	 * Return the page which will allow the user to create a site.
	 * A <code>null</code> can be returned if the page is not needed to create the site
	 * @return the page which will allow the user to create a site or <code>null</code>
	 */
	protected abstract SiteCreationWizardPage createSitePage();
	
	/**
	 * This method is called when the SiteCreationage returned by this wizard is
	 * <code>null</code>. Subclasses that return <code>null</code> as the page,
	 * must override this method to provide a non-null singleton site.
	 * @return
	 */
	protected Site getSingletonSite() {
		return null;
	}

	/* (non-Javadoc)
	 * @see org.eclipse.jface.wizard.IWizard#performFinish()
	 */
	public boolean performFinish() {
		ISiteManager siteManager = Target.getSiteManager();
		if (mainPage != null) {
			mainPage.finish(null);
			Site site = mainPage.getSite();
			if(validateSite(site, getContainer())) {
				siteManager.addSite(site);
				return true;
			} else {
				try {
					site.discarded();
					site.dispose();
				} catch (CoreException e) {
					TargetPlugin.log("Unable to purge cached login info for site being deleted.", e);
				} 
				return false;
			}
		} else {
			// There can only be one site for this wizard
			Site site = getSingletonSite();
			Assert.isNotNull(site);
			if (siteManager.getSite(site.getLocation()) == null) 
				siteManager.addSite(site);
			return true;
		}
	}

	private boolean validateSite(final Site site, IWizardContainer container) {
		try {
			container.run(true, true, new IRunnableWithProgress() {
				public void run(IProgressMonitor monitor) throws InvocationTargetException, InterruptedException {
					try {
						if (!site.canBeReached(monitor)) {
							throw new CoreException(new Status(IStatus.ERROR, TargetPlugin.ID, ITargetStatus.SITE_UNREACHABLE, NLS.bind("Site {0} is not reachable", new Object[] { site.getLocation()}), null));
						}
					} catch (CoreException e) {
						throw new InvocationTargetException(e);
					}
				}
			});
			return true;
		} catch (InvocationTargetException e) {
			if(MessageDialog.openQuestion(container.getShell(),
					"Could not Connect",
					NLS.bind("An error occurred connecting to site {0}: {1}. Keep the site anyway?", new Object[] {site.getLocation(), e.getTargetException().getMessage()}))) {
						return true;
				}
		} catch (InterruptedException e) {
			// Ignore
		}
		return false;
	}
}
