/*******************************************************************************
 * Copyright (c) 2000, 2004 IBM Corporation and others.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * Contributors:
 *     IBM Corporation - initial API and implementation
 *******************************************************************************/
package org.eclipse.team.internal.ftp.subscriber;

import java.io.*;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.HashMap;
import java.util.Map;

import org.eclipse.core.resources.IFile;
import org.eclipse.core.resources.IResource;
import org.eclipse.core.runtime.*;
import org.eclipse.ftp.*;
import org.eclipse.target.Site;
import org.eclipse.target.internal.ftp.FtpSite;
import org.eclipse.target.internal.ftp.FtpTargetResource;
import org.eclipse.team.core.Team;
import org.eclipse.team.core.TeamException;
import org.eclipse.team.core.variants.IResourceVariant;
import org.eclipse.team.core.variants.ThreeWaySynchronizer;
import org.eclipse.team.internal.ftp.*;
import org.eclipse.team.internal.ftp.Policy;
import org.eclipse.team.internal.target.*;
import org.eclipse.team.internal.target.subscriber.TargetDeploymentProvider;
import org.eclipse.team.internal.target.subscriber.TargetSubscriber;

/**
 * FTP Deployment Provider
 */
public class FTPDeploymentProvider extends TargetDeploymentProvider implements IFTPRunnableContext {

	private final String CTX_URL_KEY = "URL"; //$NON-NLS-1$
	
	private static final Map connectedThreads = new HashMap();
	
	/*
	 * Constructor required for creation by the Team deployment provider manager
	 */
	public FTPDeploymentProvider() {
		// Url will be set in restoreState
	}
	
	public FTPDeploymentProvider(URL url) {
		super(url);
	}
	
	public String getID() {
		return FTPPlugin.DEPLOYMENT_PROVIDER_ID;
	}

	protected TargetSubscriber getSubscriber() {
		TargetSubscriber subscriber = FTPPlugin.getPlugin().getSubscriber();
		return subscriber;
	}

	/* (non-Javadoc)
	 * @see org.eclipse.team.core.DeploymentProvider#saveState(org.eclipse.team.core.IMemento)
	 */
	public void saveState(IMemento memento) {
		memento.putString(CTX_URL_KEY, getUrl().toExternalForm());
		// TODO: Need to persist site (or enough info to recreate the site)
	}

	/* (non-Javadoc)
	 * @see org.eclipse.team.core.DeploymentProvider#restoreState(org.eclipse.team.core.IMemento)
	 */
	public void restoreState(IMemento memento) {
		String urlString = memento.getString(CTX_URL_KEY);
		if (urlString == null) {
			handleStateError(null);
			return;
		}
		try {
			setUrl(new URL(memento.getString(CTX_URL_KEY)));
		} catch (MalformedURLException e) {
			handleStateError(e);
		}
	}

	private void handleStateError(Throwable e) {
		// invalid state. 
		// Log it and remove the provider
		FTPPlugin.logError(Policy.bind("FTPDeploymentProvider.3", getMappedContainer().getFullPath().toString()), e); //$NON-NLS-1$
//		try {
//			Team.getDeploymentManager().unmap(getMappedContainer(), this);
//		} catch (TeamException e1) {
//			FTPPlugin.log(e1);
//		}
	}

	/**
	 * Return the URL of the FTP site that the container of this provider is mapped to.
	 * @return Returns the URL.
	 */
	public URL getRemoteURL() {
		return getUrl();
	}

	/**
	 * Return the host relative path to which this deployment provider is mapped.
	 * @return
	 */
	public String getHostRelativePath() {
		return getRemoteURL().getPath();
	}

	/* (non-Javadoc)
	 * @see org.eclipse.team.internal.ftp.client.IFTPRunnableContext#getRelativePath(java.net.URL)
	 */
	public IPath getRelativePath(URL url) {
		return UrlUtil.getTrailingPath(url, getUrl());
	}

	/* (non-Javadoc)
	 * @see org.eclipse.team.internal.ftp.client.IFTPRunnableContext#run(org.eclipse.team.internal.ftp.client.IFtpRunnable, org.eclipse.core.runtime.IProgressMonitor)
	 */
	public void run(IFtpRunnable runnable, IProgressMonitor monitor) throws TeamException {
		// Determine if we are nested or not
		boolean isOuterRun = false;
		URL baseUrl = getUrl();
		IClient client = internalGetClient(baseUrl);
		
		try {
			monitor = Policy.monitorFor(monitor);
			monitor.beginTask(null, 110 + (client.isOpen() ? 0 : 20));
			if (!client.isOpen()) {
				// The path always has a leading / to remove
				String path = baseUrl.getPath();
				if (path.length() > 0 && path.charAt(0) == '/') {
					path = path.substring(1);
				}
				client.open(Policy.subMonitorFor(monitor, 10));
				if (path.length() > 0) {
					client.changeDirectory(path, Policy.subMonitorFor(monitor, 10));
				}
				registerClient(client);
				isOuterRun = true;
			}
			client.run(runnable, Policy.subMonitorFor(monitor, 100));
		
		} catch (FtpException e) {
			throw FTPException.wrapException(e.getMessage(), e);
		} finally {
			if (isOuterRun) {
				try {
					client.close(Policy.subMonitorFor(monitor, 10));
				} catch (FtpException e) {
					FTPPlugin.log(e);
				} finally {
					deregisterClient();
				}
			}
			monitor.done();
		}
	}

	/**
	 * Return the client that is in use by the current thread
	 * or <code>null</code> if no client is available.
	 * @return a connected client
	 */
	public IClient getConnectedClient() {
		return (IClient)connectedThreads.get(Thread.currentThread());
	}
	
	public IClient registerClient(IClient client) {
		return (IClient)connectedThreads.put(Thread.currentThread(), client);
	}
	
	public IClient deregisterClient() {
		return (IClient)connectedThreads.remove(Thread.currentThread());
	}
	
	private IClient internalGetClient(URL url) throws TeamException {
		IClient client = getConnectedClient();
		if (client == null) {
			// Create client
			try {
				client = Ftp.createClient(url);
				client.setAuthentication(getUsername(url), getPassword(url));
				client.setDataTransferMode(getUsePassive(url) ? IClient.PASSIVE_DATA_TRANSFER : IClient.PORT_DATA_TRANSFER, true);
				client.setTimeout(getTimeout(url));
			} catch (FtpException e) {
				FTPException.wrapException(e.getMessage(), e);
			}
		}
		return client;
	}
	
	
	private String getPassword(URL url) throws TeamException {
		// First, attempt to get the password from the url
		String userInfo = url.getUserInfo();
		if (userInfo != null) {
			int colon = userInfo.indexOf(':');
			if (colon != -1) {
				return userInfo.substring(colon + 1);
			}
		}
		// If the above fails, try to get the password from the site
		return getFTPSite(url).getPassword();
	}

	private String getUsername(URL url) throws TeamException {
		// First, attempt to get the username from the url
		String userInfo = url.getUserInfo();
		if (userInfo != null) {
			int colon = userInfo.indexOf(':');
			if (colon == -1) {
				return userInfo;
			} else {
				return userInfo.substring(0, colon);
			}
		}
		// If the above fails, try to get the username from the site
		return getFTPSite(url).getUsername();
	}

	private boolean getUsePassive(URL url) throws TeamException {
		return getFTPSite(url).isPassive();
	}

	protected FtpSite getFTPSite(URL url) throws TeamException {
		Site site = getSite(url);
		if (site instanceof FtpSite) {
			return (FtpSite)site;
		}
		throw new TeamException(Policy.bind("FTPDeploymentProvider.4", url.toExternalForm()), null); //$NON-NLS-1$
	}

	private int getTimeout(URL url) throws TeamException {
		return getFTPSite(url).getTimeout();
	}

	/* (non-Javadoc)
	 * @see org.eclipse.team.internal.ftp.client.IFTPRunnableContext#getOpenClient()
	 */
	public IClient getOpenClient() {
		return getConnectedClient();
	}
	
	public IResourceVariant getResourceVariant(IResource resource, byte[] syncBytes) throws TeamException {
		try {
			if (syncBytes == null) return null;
			return FTPSubscriberResource.create(this, getUrl(resource), syncBytes);
		} catch (MalformedURLException e) {
			FTPPlugin.logError(Policy.bind("FTPDeploymentProvider.5", resource.getFullPath().toString()), e); //$NON-NLS-1$
			return null;
		}
	}

	/* (non-Javadoc)
	 * @see org.eclipse.team.core.target.TargetDeploymentProvider#run(org.eclipse.team.internal.core.target.ITargetRunnable, org.eclipse.core.runtime.IProgressMonitor)
	 */
	public void run(final ITargetRunnable runnable, IProgressMonitor monitor) throws TeamException {
		final TeamException[] exception = new TeamException[] { null };
		run(new IFtpRunnable() {
			public void run(IProgressMonitor monitor) throws FtpException {
				try {
					FTPDeploymentProvider.super.run(runnable, monitor);
				} catch (TeamException e) {
					exception[0] = e;
				}
			}
		}, monitor);
		if (exception[0] != null) {
			throw exception[0];
		}
	}

	/* (non-Javadoc)
	 * @see org.eclipse.team.core.target.TargetDeploymentProvider#putFile(org.eclipse.core.resources.IFile, org.eclipse.core.runtime.IProgressMonitor)
	 */
	public void putFile(final IFile localFile, IProgressMonitor monitor) throws TeamException {
		final TeamException[] exception = new TeamException[] { null };
		run(new IFtpRunnable() {
			public void run(IProgressMonitor monitor) throws FtpException {
				try {
					putFile(getOpenClient(), getRelativePath(localFile).toString(), localFile,
							Team.getType(localFile) != Team.TEXT, monitor);
				} catch (FTPException e) {
					exception[0] = e;
				}

			}
		}, monitor);
		if (exception[0] != null) {
			throw exception[0];
		}
	}
	
	/**
	 * Stores a local file on the remote system.
	 * @param filePath the absolute or relative path of the file
	 * @param localFile the local file to send
	 * @param binary if true, uses binary transfer type
	 * @param monitor the progress monitor, or null
	 */
	public void putFile(IClient client, String filePath, IFile localFile, boolean binary,
		IProgressMonitor monitor) throws FTPException {
		
		InputStream in = null;
		try {
			in = new BufferedInputStream(localFile.getContents());
			FtpTargetResource.putFile(client, filePath, in, getFileSize(localFile), binary, monitor);
		} catch (CoreException e) {
			throw FTPException.wrapException(Policy.bind("FTPClient.ErrorSendingFile"), e); //$NON-NLS-1$
		} finally {
			if (in != null)
				try {
					in.close();
				} catch (IOException e1) {
					// Ignore exceptions on close;
				}
		}
		
	}

	/* (non-Javadoc)
	 * @see org.eclipse.team.core.target.TargetDeploymentProvider#deleteRemote(org.eclipse.core.resources.IResource, org.eclipse.core.runtime.IProgressMonitor)
	 */
	public void deleteRemote(IResource resource, IProgressMonitor monitor) throws TeamException {
		try {
			if (resource.getType() == IResource.FILE) {
				getOpenClient().deleteFile(getRelativePath(resource).toString(), monitor);
			} else {
				getOpenClient().deleteDirectory(getRelativePath(resource).toString(), monitor);
			}
		} catch (FtpException e) {
			throw FTPException.wrapException(e);
		}
	}
	
	public void createRemoteDirectory(IResource resource, IProgressMonitor monitor) throws TeamException {
		try {
			try {
				getOpenClient().createDirectory(getRelativePath(resource).toString(), monitor);
			} catch (FtpException e) {
				if (e.getStatus().getCode() == IFtpStatus.DOES_NOT_EXIST) {
					// A parent must not exist, try to create it
					createRemoteDirectory(resource.getParent(), monitor);
					getOpenClient().createDirectory(getRelativePath(resource).toString(), monitor);
				} else {
					throw e;
				}
			}
		} catch (FtpException e) {
			// Some servers use 550 (DOES_NOT_EXIST) when the directory already exists.
			// Therefore, don't report an error in either case. If the directory wasn't created
			// another operation will fail anyway so we don't need to report it here
			// P.S. The repercusions of this are that server that return 550 cannot create multiple
			// ancesters for a file (i.e. upload of f1/f2/file will fail if f1 doesn't exist.
			if (e.getStatus().getCode() == IFtpStatus.DIRECTORY_EXIST || e.getStatus().getCode() == IFtpStatus.DOES_NOT_EXIST) return;
			throw FTPException.wrapException(e);
		}
	}

	/* (non-Javadoc)
	 * @see org.eclipse.team.core.target.TargetDeploymentProvider#getFile(org.eclipse.core.resources.IFile, org.eclipse.core.runtime.IProgressMonitor)
	 */
	public void getFile(IFile localFile, IProgressMonitor monitor) throws TeamException {
		getFile(getOpenClient(),getRelativePath(localFile).toString(), localFile,
				Team.getType(localFile) != Team.TEXT, false /* resume */, monitor);
	}

	/**
	 * Retrieves a remote file.
	 * @param filePath the absolute or relative path of the file
	 * @param localFile the local file to create
	 * @param binary if true, uses binary transfer type
	 * @param resume if true, attempts to resume a partial transfer, else overwrites
	 * @param monitor the progress monitor, or null
	 */
	public void getFile(IClient client, String filePath, IFile localFile,
		boolean binary, boolean resume, IProgressMonitor monitor) throws FTPException {
		
		long resumeAt = 0;
		if (resume) {
			resumeAt = getFileSize(localFile);
		}
			
		InputStream in = null;
		
		// transfer the file
		try {
			in = FtpTargetResource.getFile(client, filePath, binary, resumeAt, monitor);
			if (localFile.exists()) {
				if (resumeAt != 0) {
					// don't bother remembering the previous (incomplete) generation of the file
					localFile.appendContents(in, false /*force*/, false /*keepHistory*/, null);
				} else {
					localFile.setContents(in, false /*force*/, true /*keepHistory*/, null);
				}
			} else {
				localFile.create(in, false /*force*/, null);
			}
		} catch (CoreException e) {
			throw FTPPlugin.wrapException(e);
		} finally {
			if (in != null) {
				try {
					in.close();
				} catch (IOException e1) {
					// Ignore close exception
				}
			}
		}
	}
	
	private long getFileSize(IFile file) {
		if (! file.exists()) return 0;
		return file.getLocation().toFile().length();
	}
	
	/* (non-Javadoc)
	 * @see org.eclipse.team.core.target.TargetDeploymentProvider#getSynchronizer()
	 */
	public ThreeWaySynchronizer getSynchronizer() {
		return FTPPlugin.getPlugin().getSubscriber().getSynchronizer();
	}

	/* (non-Javadoc)
	 * @see org.eclipse.team.core.target.TargetDeploymentProvider#fetchRemoteSyncBytes(org.eclipse.core.resources.IResource, org.eclipse.core.runtime.IProgressMonitor)
	 */
	public byte[] fetchRemoteSyncBytes(IResource localResource, IProgressMonitor monitor) throws TeamException {
		throw new UnsupportedOperationException();
	}

	/* (non-Javadoc)
	 * @see org.eclipse.team.internal.target.subscriber.TargetDeploymentProvider#getSite()
	 */
	public Site getSite() {
		try {
			return getFTPSite(getUrl());
		} catch (TeamException e) {
			FTPPlugin.log(e);
			return null;
		}
	}

}
